﻿#include "character.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "interface.h"
#include "game.h"
#include "player.h"
#include "monsters.h"

std::vector<CCharacter::SChampDBData>   CCharacter::dataBase;

//------------------------------------------------------
CCharacter::CCharacter()
{
    portrait = -1;
    spell[0] = 0;
    manaRegenTimer.set(MANA_REGEN_TIMER, true);
    healthRegenTimer.set(HEALTH_REGEN_TIMER, true);
}

//------------------------------------------------------
bool CCharacter::isDead()
{
    return (stats[eStatHealth].value <= 0);
}

//------------------------------------------------------
// initialize le personnage avec les valeurs de la base de données
void CCharacter::fromDB(int num)
{
    SChampDBData&   db = dataBase[num];

    firstName = db.firstName;
    lastName = db.lastName;
    portrait = db.portrait;

    for (size_t i = 0; i < eStatCount; ++i)
    {
        stats[i].startValue = db.stats[i];
        stats[i].maxValue = db.stats[i];
        stats[i].value = db.stats[i];
    }
    stats[eStatLoad].value = 0;

    for (size_t i = 0; i < eBodyPartCount; ++i)
        bodyObjects[i].setType(0);

    for (size_t i = 0; i < db.objects.size(); ++i)
    {
        int pos = db.objects[i].pos;
        int type = db.objects[i].type;
        bodyObjects[pos].setType(type);
        objects.setDefaultParams(&bodyObjects[pos]);
    }

    updateLoad();
}

//------------------------------------------------------
void CCharacter::readCharactersDB()
{
    // tags XML dans le même ordre que EStatsNames
    static const char   statNames[][20] =
    {
        "health",
        "stamina",
        "mana",
        "strength",
        "dexterity",
        "wisdom",
        "vitality",
        "anti_magic",
        "anti_fire",
        "load"
    };

    QDomDocument doc;

    QFile f("databases/champions.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    champNode = root.firstChild();

    while(!champNode.isNull())
    {
        if (champNode.isElement())
        {
            QDomElement champion = champNode.toElement();

            if (champion.tagName() == "champion")
            {
                SChampDBData   newChamp;

                std::string firstNameId = champion.attribute("firstName").toUtf8().constData();
                newChamp.firstName = interface.getText(firstNameId);
                std::string lastNameId = champion.attribute("lastName").toUtf8().constData();
                newChamp.lastName = interface.getText(lastNameId);

                QDomElement champInfo = champion.firstChild().toElement();

                while(!champInfo.isNull())
                {
                    if (champInfo.tagName() == "portrait")
                    {
                        newChamp.portrait = champInfo.text().toInt();
                    }
                    else if (champInfo.tagName() == "object")
                    {
                        struct SDBObject    obj;

                        if (champInfo.attribute("pos") == "H")
                            obj.pos = eBodyPartHead;
                        else if (champInfo.attribute("pos") == "N")
                            obj.pos = eBodyPartNeck;
                        else if (champInfo.attribute("pos") == "T")
                            obj.pos = eBodyPartTorso;
                        else if (champInfo.attribute("pos") == "L")
                            obj.pos = eBodyPartLegs;
                        else if (champInfo.attribute("pos") == "F")
                            obj.pos = eBodyPartFeet;
                        else if (champInfo.attribute("pos") == "h")
                            obj.pos = eBodyPartRightHand;
                        else if (champInfo.attribute("pos") == "P")
                            obj.pos = eBodyPartPouch1;
                        else
                            obj.pos = champInfo.attribute("pos").toInt();

                        obj.type = champInfo.text().toInt();
                        newChamp.objects.push_back(obj);
                    }
                    else
                    {
                        for (size_t i = 0; i < eStatCount; ++i)
                        {
                            if (champInfo.tagName() == QString(statNames[i]))
                                newChamp.stats[i] = champInfo.text().toInt();
                        }
                    }
                    champInfo = champInfo.nextSibling().toElement();
                }
                newChamp.stats[eStatLoad] *= 10;
                dataBase.push_back(newChamp);
            }
        }
        champNode = champNode.nextSibling();
    }
}

//------------------------------------------------------
void CCharacter::updateLoad()
{
    int type;
    stats[eStatLoad].value = 0;

    for (int i = 0; i < eBodyPartCount; ++i)
    {
        type = bodyObjects[i].getType();
        stats[eStatLoad].value += objects.mObjectInfos[type].weight;
    }

    if (&game.characters[interface.selectedChampion] == this)
    {
        type = mouse.mObjectInHand.getType();
        stats[eStatLoad].value += objects.mObjectInfos[type].weight;
    }
}

//------------------------------------------------------
int CCharacter::overload()
{
    int value = stats[eStatLoad].value;
    int max = stats[eStatLoad].maxValue;

    if (value > max)
        return 2;
    else if (value > (625 * max) / 1000)
        return 1;
    else
        return 0;
}

//------------------------------------------------------
void CCharacter::load(FILE* handle)
{
    // charge le portrait
    fread(&portrait, sizeof(portrait), 1, handle);

    // charge le nom
    int size;
    char    c;
    fread(&size, sizeof(size), 1, handle);
    firstName.clear();
    for (int i = 0; i < size; ++i)
    {
        fread(&c, 1, 1, handle);
        firstName += c;
    }

    fread(&size, sizeof(size), 1, handle);
    lastName.clear();
    for (int i = 0; i < size; ++i)
    {
        fread(&c, 1, 1, handle);
        lastName += c;
    }

    // charge les stats
    for (int i = 0; i < eStatCount; ++i)
    {
        fread(&stats[i].value, sizeof(stats[i].value), 1, handle);
        fread(&stats[i].maxValue, sizeof(stats[i].maxValue), 1, handle);
        fread(&stats[i].startValue, sizeof(stats[i].startValue), 1, handle);
    }

    // charge les objets
    for (int i = 0; i < eBodyPartCount; ++i)
        bodyObjects[i].load(handle);

    // charge le sort courant
    fread(spell, 4, 1, handle);
    spell[4] = 0;
}

//------------------------------------------------------
void CCharacter::save(FILE* handle)
{
    // sauve le portrait
    fwrite(&portrait, sizeof(portrait), 1, handle);

    // sauve le nom
    int size = firstName.size();
    fwrite(&size, sizeof(size), 1, handle);
    fwrite(firstName.c_str(), size, 1, handle);

    size = lastName.size();
    fwrite(&size, sizeof(size), 1, handle);
    fwrite(lastName.c_str(), size, 1, handle);

    // sauve les stats
    for (int i = 0; i < eStatCount; ++i)
    {
        fwrite(&stats[i].value, sizeof(stats[i].value), 1, handle);
        fwrite(&stats[i].maxValue, sizeof(stats[i].maxValue), 1, handle);
        fwrite(&stats[i].startValue, sizeof(stats[i].startValue), 1, handle);
    }

    // sauve les objects
    for (int i = 0; i < eBodyPartCount; ++i)
        bodyObjects[i].save(handle);

    // sauve le sort courant
    fwrite(spell, 4, 1, handle);
}

//------------------------------------------------------
void CCharacter::update()
{
    if (portrait == -1)
        return;

    // met à jour les torches
    if (bodyObjects[eBodyPartLeftHand].getType() == OBJECT_TYPE_TORCH)
    {
        CTimer  t;
        t.set(bodyObjects[eBodyPartLeftHand].getIntParam("Charges"), true);
        t.update();
        bodyObjects[eBodyPartLeftHand].setIntParam("Charges", t.get());
    }

    if (bodyObjects[eBodyPartRightHand].getType() == OBJECT_TYPE_TORCH)
    {
        CTimer  t;
        t.set(bodyObjects[eBodyPartRightHand].getIntParam("Charges"), true);
        t.update();
        bodyObjects[eBodyPartRightHand].setIntParam("Charges", t.get());
    }

    // affiche les dégats
    damagesTimer.update();

    if (stats[eStatHealth].value != 0)
    {
        // régénération de mana
        if (manaRegenTimer.update() == true)
        {
            if (portrait != -1)
                if (stats[eStatMana].value < stats[eStatMana].maxValue)
                    stats[eStatMana].value++;

            if (interface.isSleeping() == false)
                manaRegenTimer.set(MANA_REGEN_TIMER, true);
            else
                manaRegenTimer.set(MANA_REGEN_TIMER_SLEEP, true);
        }

        // régénération de vie
        if (healthRegenTimer.update() == true)
        {
            if (portrait != -1)
                if (stats[eStatHealth].value < stats[eStatHealth].maxValue)
                    stats[eStatHealth].value++;

            if (interface.isSleeping() == false)
                healthRegenTimer.set(HEALTH_REGEN_TIMER, true);
            else
                healthRegenTimer.set(HEALTH_REGEN_TIMER_SLEEP, true);
        }
    }
}

//------------------------------------------------------
void CCharacter::enterSleep()
{
    manaRegenTimer.set(MANA_REGEN_TIMER_SLEEP, true);
    healthRegenTimer.set(HEALTH_REGEN_TIMER_SLEEP, true);
}

//------------------------------------------------------
void CCharacter::exitSleep()
{
    manaRegenTimer.set(MANA_REGEN_TIMER, true);
    healthRegenTimer.set(HEALTH_REGEN_TIMER, true);
}

//------------------------------------------------------
int CCharacter::getDexterity()
{
    int dexterity = stats[eStatDexterity].value + RANDOM(8);

    // diminue la dextérité en fonction de la charge
    dexterity -= (dexterity / 2) * stats[eStatLoad].value / stats[eStatLoad].maxValue;

    // si on dort...
    if (interface.isSleeping() == true)
        dexterity /= 2;

    // limite la valeur
    int min = 1 + RANDOM(8);
    int max = 100 - RANDOM(8);

    if (dexterity < min)
        dexterity = min;

    if (dexterity > max)
        dexterity = max;

    return dexterity;
}

//------------------------------------------------------
void CCharacter::hitChampion(int num, int damages)
{
    // déjà mort ou vide ?
    if (portrait == -1 || isDead() == true)
        return;

    displayedDamages = damages;
    damagesTimer.set(90, false);
    stats[eStatHealth].value -= damages;

    static CVec2    championPos[4][4] =
    {
        {{0, 0}, {1, 0}, {0, 1}, {1, 1}},   // up
        {{0, 1}, {0, 0}, {1, 1}, {1, 0}},   // left
        {{1, 1}, {0, 1}, {1, 0}, {0, 0}},   // down
        {{1, 0}, {1, 1}, {0, 0}, {0, 1}}    // right
    };

    // mort ?
    if (stats[eStatHealth].value <= 0)
    {
        stats[eStatHealth].value = 0;

        // jette tous les objet sur le sol
        CVec2   stackPos = player.pos * 2 + championPos[player.dir][num];
        CObjectStack*   stack = map.addObjectsStack(stackPos, eWallSideMax);

        for (int i = 0; i < eBodyPartCount; ++i)
        {
            CObject&    obj = bodyObjects[i];

            if (obj.getType() != 0)
            {
                stack->addObject(obj);
                obj.setType(0);
            }
        }

        // jette les os du champion
        CObject bones;
        bones.setType(OBJECT_TYPE_BONES);
        bones.setIntParam("Champion", num);
        stack->addObject(bones);
    }
}

//------------------------------------------------------
int CCharacter::getMeleeDamage(CVec2 monsterPos, CAttacks::SAttack& attack)
{
    int groupIndex = map.findMonsterGroupIndex(monsterPos);
    CMonsterGroup&  mapGroup = map.mMonsterGroups[groupIndex];
    int monsterType = mapGroup.getType();
    CMonsters::SMonsterData&    monsterData = monsters.monstersDatas[monsterType];
    int weaponType = bodyObjects[eBodyPartRightHand].getType();

    // est-ce qu'on touche ?
    int monsterDexterity = monsterData.dexterity + game.levelMultiplier() * 2 + RANDOM(32) - 16;
    int championDexterity = getDexterity();

    if (championDexterity > monsterDexterity ||
        ((RANDOM(75) + 1) <= attack.hitProbability) ||
        RANDOM(4) == 0)
    {
        int damages = 0;
        int monsterDefense = 0;

        // calcule les dégâts
        damages = getStrength(eBodyPartRightHand);

        if (damages != 0)
        {
            damages += RANDOM(damages / 2 + 1);
            damages = (damages * attack.damages) / 32;
            monsterDefense = monsterData.defense + game.levelMultiplier() * 2 + RANDOM(32);

            if (weaponType == OBJECT_TYPE_DIAMOND_EDGE)
            {
                monsterDefense -= monsterDefense / 4;
            }
            else if (weaponType == OBJECT_TYPE_HARDCLEAVE)
            {
                monsterDefense -= monsterDefense / 8;
            }

            damages = damages - monsterDefense - RANDOM(32);
        }

        // dernière chance ?
        if (damages <= 1)
        {
            int damages2 = damages;
            damages = RANDOM(4);

            if (damages != 0)
            {
                damages++;
                damages2 += RANDOM(16);

                if (damages2 > 0 || RANDOM(2))
                {
                    damages += RANDOM(4);

                    if (RANDOM(4) == 0)
                        damages += MAX(0, damages2 + RANDOM(16));
                }
            }
        }

        // randomise les dégats
        if (damages != 0)
        {
            damages /= 2;
            damages += RANDOM(damages) + RANDOM(4);
            damages += RANDOM(damages);
            damages /= 4;
            damages += RANDOM(4) + 1;
        }

        return damages;
    }
    return 0;
}

//------------------------------------------------------
int CCharacter::getStrength(int objType)
{
    int strength = stats[eStatStrength].value + RANDOM(16);
    CObjects::CObjectInfo&    objInfo = objects.mObjectInfos[objType];

    // poids de l'object
    int loadLimit = stats[eStatLoad].maxValue / 16;
    int loadLimit2 = loadLimit + (loadLimit - 12) / 2;

    if (objInfo.weight <= loadLimit)
    {
        strength += objInfo.weight - 12;
    }
    else if (objInfo.weight <= loadLimit2)
    {
        strength += (objInfo.weight - loadLimit) / 2;
    }
    else
    {
        strength -= (objInfo.weight - loadLimit2) * 2;
    }

    // dégâts de l'arme
    strength += objInfo.damages;

    // limite la valeur
    strength /= 2;

    if (strength < 0)
        strength = 0;

    if (strength > 100)
        strength = 100;

    return strength;
}
