﻿#include "game.h"
#include "tiles.h"
#include "walls.h"
#include "doors.h"
#include "player.h"
#include "../common/sources/system/fileCache.h"
#include "../common/sources/system/cgraphics2d.h"
#include <QDebug>
#include <QPainter>
#include "interface.h"
#include "bridge.h"
#include "keyboard.h"
#include "objects.h"
#include "spells.h"
#include "scripts.h"
#include "monsters.h"
#include "explosions.h"

CGame   game;

//-----------------------------------------------------------------------------------------
CGame::CGame()
{
    currentLevel = 0;
}

//-----------------------------------------------------------------------------------------
void CGame::drawWall(QImage* image, CVec2 mapPos, CVec2 tablePos, EWallSide side, bool flip)
{
    CTile*  tile = map.getTile(mapPos);
    EWallSide playerSide = player.getWallSide(side);

    //----------------------------------------------------------------------------
    if (tile->mWalls[playerSide].getType() != eWallNothing)
    {
        const char* fileName = walls.getFilename(tablePos, side, flip);

        if (fileName[0] != 0)
        {
            // dessine le mur
            QImage  wallGfx = fileCache.getImage(fileName);
            CVec2 gfxPos = walls.getGfxPos(tablePos, side, flip);

            graph2D.drawImage(image, gfxPos, wallGfx, 0, flip);

            // dessine les décors
            CWall*  wall = &tile->mWalls[playerSide];

            if (wall->getType() == eWallSimple)
            {
                //------------------------------------------------------------------------------
                // mur simple
                int ornate = wall->getOrnateParam("Decor");
                walls.drawOrnate(image, tablePos, side, ornate);
            }
            else if (wall->getType() == eWallPortrait)
            {
                //------------------------------------------------------------------------------
                // miroir de champion
                walls.drawOrnate(image, tablePos, side, ORNATE_MIRROR);

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                {
                    // si le miroir est juste en face de nous, on dessine la tête du champion
                    int champion = wall->getIntParam("Champion");
                    bool isEmpty = wall->getBoolParam("estVide");

                    if (isEmpty == false)
                    {
                        CRect   rect = interface.getChampionPortraitRect(champion);
                        QImage  portraits = fileCache.getImage("gfx/interface/ChampPortraits.png");
                        CVec2   pos(96, 68);
                        graph2D.drawImageAtlas(image, pos, portraits, rect);

                        int lastChampChosed;
                        for (lastChampChosed = 0; lastChampChosed < 4; ++lastChampChosed)
                            if (characters[lastChampChosed].portrait == -1)
                                break;

                        if (lastChampChosed != 4)
                        {
                            CRect   mouseRect(pos, CVec2(pos.x + CHAMP_PORTRAIT_WIDTH - 1, pos.y + CHAMP_PORTRAIT_HEIGHT - 1));
                            mouse.addArea(eMouseAreaG_Champion, mouseRect, eCursor_Hand, (void*)wall);
                        }
                    }
                }
            }
            else if (wall->getType() == eWallDoorSide)
            {
                //------------------------------------------------------------------------------
                // coté d'une porte
                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                {
                    QImage  doorFrame = fileCache.getImage("gfx/3DView/doors/Frame_Face_2.png");
                    graph2D.drawImage(image, CVec2(96, 33), doorFrame);
                }
            }
            else if (wall->getType() == eWallTorch)
            {
                //------------------------------------------------------------------------------
                // porte-torche
                CObjectStack*   stack = map.findObjectsStack(mapPos, playerSide);
                CRect   rect = walls.drawOrnate(image, tablePos, side, (stack != NULL ? ORNATE_TORCH_FULL : ORNATE_TORCH_EMPTY));

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                    mouse.addArea(eMouseAreaG_WallTorch, rect, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y, (void*)playerSide);
            }
            else if (wall->getType() == eWallAlcove)
            {
                //------------------------------------------------------------------------------
                // alcove
                bool isResurrecting = (mapPos == walls.resurrectWallPos && playerSide == walls.resurrectWallSide);

                int type = wall->getEnumParam("Type");
                walls.drawOrnate(image, tablePos, side, ORNATE_ARCHED_ALCOVE + type);

                if (isResurrecting == true)
                    walls.drawOrnate(image, tablePos, side, ORNATE_ALCOVE_LIGHTNING);

                if (tablePos.y > 0 && side == eWallSideUp)
                {
                    objects.drawObjectsStack(image, mapPos, playerSide, tablePos);

                    if (isResurrecting == false)
                        if (tablePos == CVec2(2, 3) && mouse.mObjectInHand.getType() != 0)
                        {
                            CRect   rectLeft(CVec2(66, 84), CVec2(157, 124));
                            mouse.addArea(eMouseAreaG_DropObject, rectLeft, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y, (void*)playerSide);
                        }
                }
            }
            else if (wall->getType() == eWallText)
            {
                //------------------------------------------------------------------------------
                // mur avec texte
                std::string text = wall->getStringParam("Texte");
                text = interface.getText(text);

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                {
                    QImage  patch = fileCache.getImage("gfx/3DView/TextPatch.png");
                    graph2D.drawImage(image, CVec2(111, 97), patch);

                    interface.drawTextMultiLines(image, CVec2(112, 74), CInterface::eFontWalls, text, QColor(), 0, 11);
                }
                else
                {
                    std::vector<std::string>    lines = interface.getTextLines(text);
                    walls.drawOrnate(image, tablePos, side, ORNATE_TEXT, lines.size());
                }
            }
            else if (wall->getType() == eWallLock)
            {
                //------------------------------------------------------------------------------
                // serrure
                int type = wall->getEnumParam("Type");
                CRect   rect = walls.drawOrnate(image, tablePos, side, ORNATE_IRON_LOCK + type);

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                    mouse.addArea(eMouseAreaG_Lock, rect, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y, (void*)playerSide);
            }
            else if (wall->getType() == eWallSwitch)
            {
                //------------------------------------------------------------------------------
                // bouton
                int     type = wall->getEnumParam("Type");
                bool    isOn = wall->getBoolParam("estOn");
                int     ornate = ORNATE_LEVER_OFF + type * 2 + isOn;
                CRect   rect = walls.drawOrnate(image, tablePos, side, ornate);

                if (tablePos == CVec2(2, 3) && side == eWallSideUp)
                    mouse.addArea(eMouseAreaG_Switch, rect, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y, (void*)playerSide);
            }
        }
    }
}

//-----------------------------------------------------------------------------------------
void CGame::displayMainView(QImage* image)
{
    if (interface.isInInventory() == false &&
        interface.isGamePaused() == false &&
        interface.isSleeping() == false)
    {
        CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
        mouse.addArea(eMouseArea_None, mainRect, eCursor_Hand);

		// dessine le sol et le plafond
        QImage  floorGfx = fileCache.getImage("gfx/3DView/Floor.png");
        QImage  ceilingGfx = fileCache.getImage("gfx/3DView/Ceiling.png");

        bool flip = (player.pos.x + player.pos.y + player.dir) & 1;

        graph2D.drawImage(image, CVec2(0, 33), ceilingGfx, 0, flip);
        graph2D.drawImage(image, CVec2(0, 72), floorGfx, 0, flip);

		// dessine les murs
        CVec2   tablePos;
        CVec2   localPos;
        CVec2   mapPos;

        for (int y = 0; y < WALL_TABLE_HEIGHT; ++y)
            for (int x = 0; x < WALL_TABLE_WIDTH; ++x)
            {
                tablePos = walls.getDrawSequence(CVec2(x, y));
                localPos = walls.getMapPos(tablePos);
                mapPos = player.getPosFromLocal(localPos);
                CTile*  tile = map.getTile(mapPos);

                if (tile != NULL)
                {
                    // dessine les murs
                    drawWall(image, mapPos, tablePos, eWallSideUp, flip);
                    drawWall(image, mapPos, tablePos, eWallSideLeft, flip);
                    drawWall(image, mapPos, tablePos, eWallSideRight, flip);

                    // dessine les escaliers
                    if (tile->getType() == eTileStairs)
                        tiles.drawStairs(image, tile, tablePos);

                    // dessine les trous
                    if (tile->getType() == eTilePit && tile->getBoolParam("estFerme") == false)
                        tiles.drawPit(image, tablePos);

                    // dessine les décors au sol
                    if (tile->getType() == eTileGround)
                    {
                        // case de sol
                        int ornate = tile->getOrnateParam("Decor");
                        tiles.drawFloorOrnate(image, tablePos, ornate);
                    }
                    else if (tile->getType() == eTilePressPlate)
                    {
                        // plaque de pression
                        int ornate = tile->getEnumParam("Type");

                        if (ornate != 0)
                            tiles.drawFloorOrnate(image, tablePos, ornate + 6);
                    }

                    // dessine les objets de la ligne arrière
                    objects.drawBackRow(image, mapPos, tablePos);

                    // dessine les portes
                    if (tile->getType() == eTileDoor)
                        doors.draw(image, mapPos, tablePos);

                    // dessine les objets de la ligne avant
                    objects.drawFrontRow(image, mapPos, tablePos);

                    // dessine les ennemis
                    monsters.draw(image, mapPos, tablePos);

                    // dessine les explosions
                    explosions.draw(image, mapPos, tablePos);

                    // dessine les teleporteurs
                    if (tile->getType() == eTileTeleporter)
                        tiles.drawTeleporter(image, tablePos);
                }
            }

        // lumière
        if (currentLevel != 1)
        {
            float   level = getLightLevel();

            if (level != MAX_LIGHT_LEVEL)
                graph2D.darkenRect(image, 1.0f - level, mainRect);
        }
    }
}

//-----------------------------------------------------------------------------------------
float CGame::getLightLevel()
{
    int power = 0;

    // gère la puissance des torches dans les amins des champions
    int torches[8];
    int numTorches = 0;
    CObject*    obj;

    for (int i = 0; i < 4; ++i)
    {
        CCharacter* c = &game.characters[i];

        obj = &c->bodyObjects[CCharacter::eBodyPartLeftHand];
        if (obj->getType() == OBJECT_TYPE_TORCH)
            torches[numTorches++] = obj->getIntParam("Charges");

        obj = &c->bodyObjects[CCharacter::eBodyPartRightHand];
        if (obj->getType() == OBJECT_TYPE_TORCH)
            torches[numTorches++] = obj->getIntParam("Charges");
    }

    // trie les torches
    for (int i = 0; i < numTorches - 1; ++i)
        for (int j = i + 1; j < numTorches; ++j)
            if (torches[i] < torches[j])
            {
                int temp = torches[i];
                torches[i] = torches[j];
                torches[j] = temp;
            }

    // prend les 5 torches les plus puissantes (comme dans le jeu d'origine)
    for (int i = 0; i < MIN(numTorches, 5); ++i)
        power += (torches[i] >> i) / 545;

    // sorts
    for (size_t i = 0; i < spells.mActiveSpells.size(); ++i)
    {
        CSpell* s = &spells.mActiveSpells[i];

        if (s->spell == eSpell_MagicTorch ||
            s->spell == eSpell_Light ||
            s->spell == eSpell_Darkness)
        {
            power += s->getPower();
        }
    }

    // calcule le niveau global de lumière
    float   level = ((float)power / 256.0f) * (MAX_LIGHT_LEVEL - MIN_LIGHT_LEVEL) + MIN_LIGHT_LEVEL;

    if (level > MAX_LIGHT_LEVEL)
        level = MAX_LIGHT_LEVEL;

    return level;
}

//-----------------------------------------------------------------------------------------
QImage& CGame::mainLoop()
{
    static QImage image(SCREEN_WIDTH,
                        SCREEN_HEIGHT,
                        QImage::Format_RGBA8888);

    // évènements spéciaux
    if (player.shouldTakeStairs)
        player.takeStairs();

    if (player.shouldTeleport)
        player.teleport();

    if (player.shouldFall && sound->isFinished(player.pos, "sound/Scream.wav") == true)
        player.fall();

    if (interface.shouldLoad == true)
    {
        loadGame(interface.saveName.c_str());
        interface.shouldLoad = false;
        interface.setMainState(CInterface::eMainLoadedDialog);
    }

    if (interface.shouldSave == true)
    {
        saveGame(interface.saveName.c_str());
        interface.shouldSave = false;
        interface.setMainState(CInterface::eMainSavedDialog);
    }

    // efface les zones souris
    mouse.clearAreas();

    // efface l'écran
    image.fill(BLACK);

    // affichage
    displayMainView(&image);
    interface.display(&image);

    // logique du jeu
    mouse.update();
    SMouseArea* clickedArea = mouse.clickedArea(&image);

    keyboard.update();
    update(clickedArea);
    interface.update(clickedArea);
    spells.update();
    monsters.update();
    walls.update();
    explosions.update();

    for (int i = 0; i < 4; ++i)
        if (interface.isResurrectingCharacter(i) == false)
            characters[i].update();

    return image;
}

//-----------------------------------------------------------------------------------------
void CGame::update(SMouseArea* clickedArea)
{
    frameCounter++;
    doors.update();
    tiles.updatePressPlates();

    if (clickedArea != NULL)
    {
        if (mouse.mButtonPressing == true)
        {
            if (clickedArea->type == eMouseAreaG_Champion)
            {
                lastMirrorClicked = (CWall*)clickedArea->param1;

                int lastChampChosed;
                for (lastChampChosed = 0; lastChampChosed < 4; ++lastChampChosed)
                    if (characters[lastChampChosed].portrait == -1)
                        break;

                characters[lastChampChosed].fromDB(lastMirrorClicked->getIntParam("Champion"));
                interface.enterResurrect(lastChampChosed);
            }
            else if (clickedArea->type == eMouseAreaG_DoorButton)
            {
                CVec2   mapPos;
                mapPos.x = (int)clickedArea->param1;
                mapPos.y = (int)clickedArea->param2;
                CTile*  tile = map.getTile(mapPos);
                bool isOpened = tile->getBoolParam("estOuverte");
                tile->setBoolParam("estOuverte", !isOpened);
                sound->play(mapPos, "sound/Switch.wav");
            }
            else if (clickedArea->type == eMouseAreaG_PickObject)
            {
                CObjectStack*   stack = (CObjectStack*)clickedArea->param1;
                int index = (int)clickedArea->param2;
                mouse.mObjectInHand = stack->getObject(index);
                stack->removeObject(index);

                if (stack->getSize() == 0)
                    map.removeObjectsStack(stack->mPos, stack->mSide);

                characters[interface.selectedChampion].updateLoad();
            }
            else if (clickedArea->type == eMouseAreaG_DropObject)
            {
                // récupère la position et le coté
                CVec2       pos = CVec2((int)clickedArea->param1, (int)clickedArea->param2);
                EWallSide   side = (EWallSide)(int)clickedArea->param3;

                // jette l'objet
                CObjectStack*   stack = map.addObjectsStack(pos, side);
                stack->addObject(mouse.mObjectInHand);

                // test si on ressuscite un champion
                if (mouse.mObjectInHand.getType() == OBJECT_TYPE_BONES &&
                    side != eWallSideMax)
                {
                    CWall&  wall = map.getTile(pos)->mWalls[side];

                    if (wall.getType() == eWallAlcove &&
                        wall.getEnumParam("Type") == 2)
                    {
                        // ressuscite
                        sound->play(pos, "sound/Spell.wav");
                        walls.resurrectWallPos = pos;
                        walls.resurrectWallSide = side;
                        walls.resurrectTimer.set(120, true);
                    }
                }

                mouse.mObjectInHand.setType(0);
                characters[interface.selectedChampion].updateLoad();

                if (side == eWallSideMax)
                {
                    CVec2   mapPos = pos / 2;
                    CTile*  tile = map.getTile(mapPos);

                    if (tile->getType() == eTileTeleporter && tile->getBoolParam("pourObjets") == true)
                        teleportStacks(mapPos);

                    if (tile->getType() == eTilePit && tile->getBoolParam("estFerme") == false)
                        fallStacks(mapPos);
                }
            }
            else if (clickedArea->type == eMouseAreaG_WallTorch)
            {
                CVec2   pos;
                EWallSide   side;

                pos.x = (int)clickedArea->param1;
                pos.y = (int)clickedArea->param2;
                side = (EWallSide)((int)clickedArea->param3);

                int type = mouse.mObjectInHand.getType();
                CObjectStack* stack = map.findObjectsStack(pos, side);

                if (type == 0 && stack != NULL)
                {
                    mouse.mObjectInHand = stack->getObject(0);
                    map.removeObjectsStack(pos, side);
                }
                else if (type == OBJECT_TYPE_TORCH && stack == NULL)
                {
                    stack = map.addObjectsStack(pos, side);
                    stack->addObject(mouse.mObjectInHand);
                    mouse.mObjectInHand.setType(0);
                }
                characters[interface.selectedChampion].updateLoad();
            }
            else if (clickedArea->type == eMouseAreaG_Lock)
            {
                CVec2   pos;
                EWallSide   side;

                pos.x = (int)clickedArea->param1;
                pos.y = (int)clickedArea->param2;
                side = (EWallSide)((int)clickedArea->param3);

                CObjectStack* stack = map.findObjectsStack(pos, side);

                for (size_t i = 0; i < stack->getSize(); ++i)
                    if (stack->getObject(i).getType() == mouse.mObjectInHand.getType())
                    {
                        stack->removeObject(i);
                        mouse.mObjectInHand.setType(0);
                        sound->play(pos, "sound/Switch.wav");
                        break;
                    }

                if (stack->getSize() == 0)
                {
                    CWall*  wall = &map.getTile(pos)->mWalls[side];
                    std::string script = wall->getScriptParam("siOuvert");
                    scripts.execute(script);
                }
            }
            else if (clickedArea->type == eMouseAreaG_Switch)
            {
                CVec2   pos;
                EWallSide   side;

                pos.x = (int)clickedArea->param1;
                pos.y = (int)clickedArea->param2;
                side = (EWallSide)((int)clickedArea->param3);
                CWall*  wall = &map.getTile(pos)->mWalls[side];

                bool    isOn = !wall->getBoolParam("estOn");
                wall->setBoolParam("estOn", isOn);
                sound->play(pos, "sound/Switch.wav");

                std::string script;

                if (isOn == false)
                    script = wall->getScriptParam("siOff");
                else
                    script = wall->getScriptParam("siOn");

                if (script.empty() == false)
                    scripts.execute(script);
            }
        }
    }
}

//-----------------------------------------------------------------------------------------
void CGame::loadLevel(int level, CVec2 pos, int side)
{
    // mémorise le niveau courant
    snapshotLevel();
    currentLevel = level;

    // cherche si le nouveau niveau est mémorisé...
    for (size_t i = 0; i < mVisitedLevels.size(); ++i)
    {
        CSnapshot*  s = &mVisitedLevels[i];

        if (s->level == level)
        {
            map.freeMemory();
            map = s->map;
            tiles.setPressPlates(s->pressPlateStates);
            player.pos = pos;
            player.dir = side;
            return;
        }
    }

    // ...sinon, charge le nouveau niveau
    static char fileName[256];
    sprintf(fileName, "maps/level%02d.map", level);
    map.load(fileName);

    player.pos = pos;
    player.dir = side;
    tiles.initMap();
    walls.init();
    monsters.init();
}

//-----------------------------------------------------------------------------------------
int CGame::levelMultiplier()
{
    static const int multipliers[] =
    {
        0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 5, 5, 6, 6
    };
    return multipliers[currentLevel - 1];
}

//-----------------------------------------------------------------------------------------
CSnapshot::CSnapshot()
{
    pressPlateStates = NULL;
}

//-----------------------------------------------------------------------------------------
CSnapshot::CSnapshot(const CSnapshot& rhs)
{
    pressPlateStates = NULL;
    *this = rhs;
}

//-----------------------------------------------------------------------------------------
CSnapshot::~CSnapshot()
{
    freePressPlates();
}

//-----------------------------------------------------------------------------------------
void CSnapshot::freePressPlates()
{
    if (pressPlateStates != NULL)
        delete[] pressPlateStates;
}

//-----------------------------------------------------------------------------------------
void CSnapshot::copyPressPlates(bool* src)
{
    freePressPlates();
    int mapSize = map.mSize.x * map.mSize.y;

    pressPlateStates = new bool[mapSize];
    for (int i = 0; i < mapSize; ++i)
        pressPlateStates[i] = src[i];
}

//-----------------------------------------------------------------------------------------
CSnapshot&  CSnapshot::operator=(const CSnapshot& rhs)
{
    level = rhs.level;
    map = rhs.map;
    copyPressPlates(rhs.pressPlateStates);
    return *this;
}

//-----------------------------------------------------------------------------------------
void CGame::snapshotLevel()
{
    if (currentLevel != 0)
    {
        // Si le niveau est déjà mémorisé, on l'efface
        std::vector<CSnapshot>::iterator it;

        for (it = mVisitedLevels.begin(); it != mVisitedLevels.end(); ++it)
            if (it->level == currentLevel)
            {
                mVisitedLevels.erase(it);
                break;
            }

        // ajoute une nouvelle entrée
        CSnapshot   s;
        s.level = currentLevel;
        s.map = map;
        s.copyPressPlates(tiles.getPressPlates());
        mVisitedLevels.push_back(s);
    }
}

//---------------------------------------------------------------------------------------------
void CGame::saveGame(const char *fileName)
{
    snapshotLevel();
    FILE*       handle = fopen(fileName, "wb");

    if (handle != NULL)
    {
        // sauve le niveau courant et la position du joueur
        fwrite(&currentLevel, sizeof(currentLevel), 1, handle);
        fwrite(&player.pos.x, sizeof(player.pos.x), 1, handle);
        fwrite(&player.pos.y, sizeof(player.pos.y), 1, handle);
        fwrite(&player.dir, sizeof(player.dir), 1, handle);

        // sauve les personnages
        for (int i = 0; i < 4; ++i)
            game.characters[i].save(handle);

        // sauve les images des maps
        int numSnapshots = mVisitedLevels.size();
        fwrite(&numSnapshots, sizeof(numSnapshots), 1, handle);

        for (int i = 0; i < numSnapshots; ++i)
            mVisitedLevels[i].save(handle);

        // sauve les sorts
        spells.save(handle);

        fclose(handle);
    }
}

//---------------------------------------------------------------------------------------------
void CGame::loadGame(const char* fileName)
{
    FILE*       handle = fopen(fileName, "rb");

    if (handle != NULL)
    {
        // charge le niveau courant et la position du joueur dans des variables temporaires car elles seront positionnées par loadLevel()
        int currentFileLevel;
        CVec2   filePlayerPos;
        uint8_t filePlayerDir;
        fread(&currentFileLevel, sizeof(currentFileLevel), 1, handle);
        fread(&filePlayerPos.x, sizeof(filePlayerPos.x), 1, handle);
        fread(&filePlayerPos.y, sizeof(filePlayerPos.y), 1, handle);
        fread(&filePlayerDir, sizeof(filePlayerDir), 1, handle);

        // charge les personnages
        for (int i = 0; i < 4; ++i)
            game.characters[i].load(handle);

        // charge les images des maps
        mVisitedLevels.clear();
        int numSnapshots;
        fread(&numSnapshots, sizeof(numSnapshots), 1, handle);
        mVisitedLevels.resize(numSnapshots);

        for (int i = 0; i < numSnapshots; ++i)
            mVisitedLevels[i].load(handle);

        // charge les sorts
        spells.load(handle);

        fclose(handle);

        currentLevel = 0; // évite de refaire une image
        loadLevel(currentFileLevel, filePlayerPos, filePlayerDir);
    }
}

//---------------------------------------------------------------------------------------------
void    CSnapshot::load(FILE* handle)
{
    fread(&level, sizeof(level), 1, handle);
    map.load(handle);

    uint8_t state;
    freePressPlates();
    int mapSize = map.mSize.x * map.mSize.y;
    pressPlateStates = new bool[mapSize];

    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
    {
        fread(&state, sizeof(uint8_t), 1, handle);
        pressPlateStates[i] = state;
    }
}

//---------------------------------------------------------------------------------------------
void    CSnapshot::save(FILE* handle)
{
    fwrite(&level, sizeof(level), 1, handle);
    map.save(handle);

    uint8_t state;
    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
    {
        state = pressPlateStates[i];
        fwrite(&state, sizeof(uint8_t), 1, handle);
    }
}

//---------------------------------------------------------------------------------------------
void CGame::teleportStacks(CVec2 mapPos)
{
    CTile*  tile = map.getTile(mapPos);
    std::vector<CObjectStack>   copyStacks;

    // copie les tas
    for (int y = 0; y < 2; ++y)
        for (int x = 0; x < 2; ++x)
        {
            CVec2 objPos = mapPos * 2 + CVec2(x, y);
            CObjectStack*    stack = map.findObjectsStack(objPos, eWallSideMax);

            if (stack != NULL)
            {
                CObjectStack    copy = *stack;
                copy.mPos = CVec2(x, y);
                copyStacks.push_back(copy);
                map.removeObjectsStack(objPos, eWallSideMax);
            }
        }

    int     lastLevel = currentLevel;
    CVec2   lastPos = player.pos;
    uint8_t lastDir = player.dir;

    int     level = tile->getIntParam("Niveau");
    CVec2   newPos = CVec2(tile->getIntParam("X"), tile->getIntParam("Y"));

    // charge le niveau de destination
    if (level != currentLevel)
        loadLevel(level, CVec2(), 0);

    // mets les tas à la nouvelle position
    for (size_t i = 0; i < copyStacks.size(); ++i)
    {
        CVec2 objPos = newPos * 2 + copyStacks[i].mPos;
        CObjectStack*   stack = map.addObjectsStack(objPos, eWallSideMax);

        for (size_t j = 0; j < copyStacks[i].getSize(); ++j)
            stack->addObject(copyStacks[i].getObject(j));
    }

    // recharge le dernier niveau
    if (level != lastLevel)
        loadLevel(lastLevel, lastPos, lastDir);
}

//---------------------------------------------------------------------------------------------
void CGame::fallStacks(CVec2 mapPos)
{
    CTile*  tile = map.getTile(mapPos);
    std::vector<CObjectStack>   copyStacks;

    // copie les tas
    for (int y = 0; y < 2; ++y)
        for (int x = 0; x < 2; ++x)
        {
            CVec2 objPos = mapPos * 2 + CVec2(x, y);
            CObjectStack*    stack = map.findObjectsStack(objPos, eWallSideMax);

            if (stack != NULL)
            {
                CObjectStack    copy = *stack;
                copy.mPos = CVec2(x, y);
                copyStacks.push_back(copy);
                map.removeObjectsStack(objPos, eWallSideMax);
            }
        }

    int     lastLevel = currentLevel;
    CVec2   lastPos = player.pos;
    uint8_t lastDir = player.dir;

    int     level = tile->getIntParam("Niveau");
    CVec2   newPos = CVec2(tile->getIntParam("X"), tile->getIntParam("Y"));

    // charge le niveau de destination
    loadLevel(level, CVec2(), 0);

    // mets les tas à la nouvelle position
    for (size_t i = 0; i < copyStacks.size(); ++i)
    {
        CVec2 objPos = newPos * 2 + copyStacks[i].mPos;
        CObjectStack*   stack = map.addObjectsStack(objPos, eWallSideMax);

        for (size_t j = 0; j < copyStacks[i].getSize(); ++j)
            stack->addObject(copyStacks[i].getObject(j));
    }

    // recharge le dernier niveau
    loadLevel(lastLevel, lastPos, lastDir);
}
