﻿#include "monsters.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "../common/sources/Map.h"
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "player.h"
#include "walls.h"
#include "interface.h"
#include "game.h"
#include "sound.h"

CMonsters    monsters;

//---------------------------------------------------------------------------------------------
CMonsters::CMonsters()
{

}

//---------------------------------------------------------------------------------------------
void CMonsters::readMonstersDB()
{
    QDomDocument doc;

    QFile f("databases/monsters.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    monsterNode = root.firstChild();

    while(!monsterNode.isNull())
    {
        if (monsterNode.isElement())
        {
            QDomElement monster = monsterNode.toElement();
            std::vector<CParamType> paramsTypesList;

            if (monster.tagName() == "monster")
            {
                QDomElement monsterInfo = monster.firstChild().toElement();
                SMonsterData    monsterData;

                while(!monsterInfo.isNull())
                {
                    if (monsterInfo.tagName() == "img_front")
                    {
                        monsterData.frontImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_side")
                    {
                        monsterData.sideImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_back")
                    {
                        monsterData.backImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_attack")
                    {
                        monsterData.attackImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "time_between_atk")
                    {
                        monsterData.timeBetweenAttack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_display_time")
                    {
                        monsterData.attackDisplayTime = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_sound")
                    {
                        monsterData.attackSound = QString("sound/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "defense")
                    {
                        monsterData.defense = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "health")
                    {
                        monsterData.baseHealth = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "attack")
                    {
                        monsterData.attack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "dexterity")
                    {
                        monsterData.dexterity = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "param")
                    {
                        QString paramType = monsterInfo.attribute("type");

                        CParamType  newParamType;

                        if (paramType == "int")
                        {
                            newParamType.mType = eParamInt;
                            newParamType.mName = monsterInfo.text();
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "enum")
                        {
                            newParamType.mType = eParamEnum;
                            newParamType.mName = monsterInfo.text();
                            newParamType.mValues = monsterInfo.attribute("values").split(";");
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "stack")
                        {
                            newParamType.mType = eParamStack;
                            paramsTypesList.push_back(newParamType);
                        }
                    }

                    monsterInfo = monsterInfo.nextSibling().toElement();
                }

                CMap::mMonstersParams.push_back(paramsTypesList);
                monstersDatas.push_back(monsterData);
            }
        }
        monsterNode = monsterNode.nextSibling();
    }
}

//---------------------------------------------------------------------------------------------
int CMonsters::getStartingLife(int type)
{
    int baseHealth = monstersDatas[type].baseHealth;

    return baseHealth * game.levelMultiplier() + RANDOM((baseHealth / 4) + 1);
}

//---------------------------------------------------------------------------------------------
void CMonsters::init()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2	group;
        int nbMonsters = mapGroup.getIntParam("nbMonstres");

        group.dir = (EMonsterDir)mapGroup.getEnumParam("Dir");

        for (int j = 0; j < 4; j++)
        {
            group.mMonsters[j].dir = group.dir;
            group.mMonsters[j].pos = eMonsterPosNone;
            group.mMonsters[j].life = 0;
        }

        if (nbMonsters == 1)
        {
            group.mMonsters[0].pos = eMonsterPosCenter;
            group.mMonsters[0].life = getStartingLife(mapGroup.getType());
        }
        else
        {
            // positions des monstres pour une direction donnée
            static const EMonsterPos startingPos[4][4] =
            {
                {eMonsterPosNorthWest, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosSouthEast},   // up
                {eMonsterPosSouthEast, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosNorthWest},   // down
                {eMonsterPosNorthWest, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosSouthEast},   // left
                {eMonsterPosSouthEast, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosNorthWest}    // right
            };

            for (int j = 0; j < 4; j++)
            {
                if (j < nbMonsters)
                {
                    group.mMonsters[j].pos = startingPos[group.dir][j];
                    group.mMonsters[j].life = getStartingLife(mapGroup.getType());
                }
            }
        }

        monsterGroups.push_back(group);
    }
}

//---------------------------------------------------------------------------------------------
CVec2 CMonsters::getMonsterPos(float x, float y)
{
    CVec2   pos;
    pos.x = 250 * (x - 4.5) / (8.5 - y) + 112;
    pos.y = 310 / (8.5 - 0.9 * y) + 66;
    return pos;
}

//---------------------------------------------------------------------------------------------
void CMonsters::draw(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    if (tablePos.y == WALL_TABLE_HEIGHT - 1)
        return;

    int groupIndex = map.findMonsterGroupIndex(mapPos);

    if (groupIndex != -1)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[groupIndex];
        CMonsterGroup2& group = monsterGroups[groupIndex];

        // position des monstres en fonction de la position du joueur
        static const float monsterPosDir[4][5][2] =
        {
            //   NW          NE          SW          SE           C
            {{0.0, 0.0}, {1.0, 0.0}, {0.0, 1.0}, {1.0, 1.0},  {0.5, 0.5}},   // haut
            {{1.0, 0.0}, {1.0, 1.0}, {0.0, 0.0}, {0.0, 1.0},  {0.5, 0.5}},   // gauche
            {{1.0, 1.0}, {0.0, 1.0}, {1.0, 0.0}, {0.0, 0.0},  {0.5, 0.5}},   // bas
            {{0.0, 1.0}, {0.0, 0.0}, {1.0, 1.0}, {1.0, 0.0},  {0.5, 0.5}}    // droite
        };

        // direction des monstres en fonction de la direction du joueur
        static const EMonsterDir monsterDirDir[4][4] =
        {
            {eMonsterDirUp,    eMonsterDirDown,  eMonsterDirLeft,  eMonsterDirRight}, // haut
            {eMonsterDirRight, eMonsterDirLeft,  eMonsterDirUp,    eMonsterDirDown},  // gauche
            {eMonsterDirDown,  eMonsterDirUp,    eMonsterDirRight, eMonsterDirLeft},  // bas
            {eMonsterDirLeft,  eMonsterDirRight, eMonsterDirDown,  eMonsterDirUp}     // droite
        };

        // on dessine la ligne arrière des monstres avant ceux de devant
        for (int row = 0; row < 3; ++row)
        {
            for (int i = 0; i < 4; ++i)
            {
                CMonster&   monster = group.mMonsters[i];

                if (monster.pos != eMonsterPosNone)
                {
                    // direction relative
                    EMonsterDir tableDir = monsterDirDir[player.dir][monster.dir];

                    // sprite
                    int type = mapGroup.getType();
                    QString spriteName;
                    bool    flip = false;

                    if (tableDir == eMonsterDirDown)
                    {
                        // de face
                        if (monster.displayAttackTimer.isRunning())
                        {
                            // attaque
                            spriteName = monstersDatas[type].attackImage;
                            if (spriteName.isEmpty() == true)
                                spriteName = monstersDatas[type].frontImage;
                        }
                        else
                        {
                            // normal
                            spriteName = monstersDatas[type].frontImage;
                        }
                    }
                    else if (tableDir == eMonsterDirUp)
                    {
                        // de dos
                        spriteName = monstersDatas[type].backImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                    }
                    else
                    {
                        // de coté
                        spriteName = monstersDatas[type].sideImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                        else if (tableDir == eMonsterDirLeft)
                            flip = true;
                    }

                    if (monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1] == (float)row / 2.0)
                    {
                        // position dans la table
                        float tablePosX = tablePos.x * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][0];
                        float tablePosY = tablePos.y * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1];

                        // position à l'écran
                        QImage  sprite = fileCache.getImage(spriteName.toStdString());
                        CVec2   pos = getMonsterPos(tablePosX, tablePosY);

                        // calcule la taille par rapport à la position de référence (la plus proche)
                        CVec2   pos0, pos1;
                        float   scale;

                        pos0 = getMonsterPos(WALL_TABLE_WIDTH, (WALL_TABLE_HEIGHT - 1) * 2 + 1);
                        pos1 = getMonsterPos(WALL_TABLE_WIDTH, tablePosY);
                        scale = 2.37 * (float)(pos1.x - 112) / (float)(pos0.x - 112);

                        // redimensionne le monstre dans une image temporaire
                        QImage  scaledMonster(sprite.size() * scale, QImage::Format_ARGB32);
                        scaledMonster.fill(TRANSPARENT);
                        graph2D.drawImageScaled(&scaledMonster, CVec2(), sprite, scale);

                        // assombrit le monstre en fonction de la distance
                        float shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePosY - 1) * 0.13f;
                        graph2D.darken(&scaledMonster, shadow);

                        // dessine le monstre (position relative à ses pieds)
                        pos.x -= scaledMonster.width() / 2;
                        pos.y -= scaledMonster.height() - 9 * scale;
                        graph2D.drawImage(image, pos, scaledMonster, 0, flip, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));
                    }
                }
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsters::update()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2& group = monsterGroups[i];

        group.update(mapGroup.mPos, mapGroup.getType());
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::update(CMonsterGroup2& group, CVec2 mapPos, uint8_t type, bool isInFront)
{
    displayAttackTimer.update();

    switch (group.state)
    {
        case eMonsterState_Wandering:
            wander();
            break;

        case eMonsterState_Fight:
            fight(mapPos, type, isInFront);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::wander()
{

}

//---------------------------------------------------------------------------------------------
void CMonster::initNextAttack(uint8_t type)
{
    int minTime = monsters.monstersDatas[type].timeBetweenAttack;
    int nextTime = minTime + RANDOM(40) - 10;

    if (minTime > 150)
        nextTime += RANDOM(80) - 20;

    nextAttackTimer.set(nextTime, true);
}

//---------------------------------------------------------------------------------------------
// choisit le champion que ce monstre va attaquer
uint8_t CMonster::chooseTarget(CVec2 mapPos)
{
    // liste les cibles possibles
    uint8_t possibleTargets[2];
    int nbTargets = 0;

    for (int i = 0; i < 4; ++i)
    {
        if (interface.isChampionEmpty(i) == false &&
            interface.isChampionDead(i) == false &&
            player.isChampionInFront(mapPos, i) == true)
        {
            possibleTargets[nbTargets++] = i;
        }
    }

    // choisit une cible
    return possibleTargets[RANDOM(nbTargets)];
}

//---------------------------------------------------------------------------------------------
void CMonster::fight(CVec2 mapPos, uint8_t type, bool isInFront)
{
    if (nextAttackTimer.update() == true)
    {
        // si le monstre est dans la ligne avant du groupe
        if (isInFront == true)
        {
            // joue le son d'attaque
            QString soundName = monsters.monstersDatas[type].attackSound;

            if (soundName.isEmpty() == false)
                sound->play(mapPos, soundName.toLocal8Bit().constData());

            // affiche l'animation du monstre
            int attackTime = monsters.monstersDatas[type].attackDisplayTime;
            displayAttackTimer.set(attackTime, true);

            // choisit un champion cible
            int target = chooseTarget(mapPos);

            // calcule les dextérités du monstre et du champion
            int monsterDexterity = monsters.monstersDatas[type].dexterity + game.levelMultiplier() * 2 + RANDOM(32) - 16;
            int championDexterity = game.characters[target].getDexterity();

            // teste si on a touché
            if (interface.isSleeping() == true ||
                monsterDexterity > championDexterity ||
                RANDOM(4) == 0)
            {
                // calcule les dégats (ici, on devrait avoir une chance de parer)
                int damages = monsters.monstersDatas[type].attack + game.levelMultiplier() * 2 + RANDOM(16);

                // randomise les dégats (d'une façon assez compliquée...)
                damages /= 2;
                damages += RANDOM(damages) + RANDOM(4);
                damages += RANDOM(damages);
                damages /= 4;
                damages += RANDOM(4) + 1;

                if (RANDOM(2))
                    damages -= RANDOM(damages / 2 + 1) - 1;

                // frappe le champion
                static char soundName[256];
                sprintf(soundName, "sound/Champion_Damaged_%d.wav", RANDOM(4) + 1);
                sound->play(player.pos, soundName);
                game.characters[target].hitChampion(target, damages);
            }

            // si on dromait, on se réveille
            if (interface.isSleeping() == true)
                interface.wakeUp();
        }
        initNextAttack(type);
    }
}

//---------------------------------------------------------------------------------------------
CMonsterGroup2::CMonsterGroup2()
{
    state = eMonsterState_Wandering;
}

//---------------------------------------------------------------------------------------------
// retourne le rang d'un monstre en fonction de la position du joueur (0 = avant, 1 = centre, 2 = arrière)
uint8_t CMonsterGroup2::getMonsterRow(CVec2 mapPos, EMonsterPos monsterPos)
{
    static const uint8_t monsterRow[4][5] =
    {
        // NW NE SW SE C
        {  0, 0, 2, 2, 1},   // haut
        {  0, 2, 0, 2, 1},   // gauche
        {  2, 2, 0, 0, 1},   // bas
        {  2, 0, 2, 0, 1}    // droite
    };

    int playerPosFromMonster;

    if (player.pos.x < mapPos.x)
        playerPosFromMonster = 1; // gauche
    else if (player.pos.x > mapPos.x)
        playerPosFromMonster = 3; // droite
    else if (player.pos.y < mapPos.y)
        playerPosFromMonster = 0; // haut
    else
        playerPosFromMonster = 2; // bas

    return monsterRow[playerPosFromMonster][monsterPos - eMonsterPosNorthWest];
}

//---------------------------------------------------------------------------------------------
// retourne true si un monstre est en première ligne de son groupe, en prenant en compte les morts
bool CMonsterGroup2::isMonsterInFront(CVec2 mapPos, EMonsterPos monsterPos)
{
    bool    isFrontRowRowAlive = false;

    for (int i = 0; i < 4; ++i)
    {
        if (mMonsters[i].pos != eMonsterPosNone &&
            getMonsterRow(mapPos, mMonsters[i].pos) < 2)
        {
            isFrontRowRowAlive = true;
            break;
        }
    }

    uint8_t row = getMonsterRow(mapPos, monsterPos);
    return (isFrontRowRowAlive == true && row < 2) ||
           (isFrontRowRowAlive == false && row == 2);
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::update(CVec2 mapPos, uint8_t type)
{
    // met à jour tous les monstres du groupe
    for (int i = 0; i < 4; ++i)
        if (mMonsters[i].pos != eMonsterPosNone)
        {
            bool    isInFront = isMonsterInFront(mapPos, mMonsters[i].pos);
            mMonsters[i].update(*this, mapPos, type, isInFront);
        }

    // gère le comportement du groupe pour chaque état
    switch (state)
    {
        case eMonsterState_Wandering:
            wander(mapPos, type);
            break;

        case eMonsterState_Fight:
            fight(mapPos, type);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::wander(CVec2 mapPos, uint8_t type)
{
    // transition en état de combat ?
    CVec2   dist = player.pos - mapPos;
    bool    isNearPlayer = (dist.x == 0 && ABS(dist.y) == 1) || (dist.y == 0 && ABS(dist.x) == 1);

    if (isNearPlayer == true)
    {
        state = eMonsterState_Fight;

        // direction vers le joueur
        EMonsterDir dir;

        if (dist.x > 0)
            dir = eMonsterDirRight;
        else if (dist.x < 0)
            dir = eMonsterDirLeft;
        else if (dist.y > 0)
            dir = eMonsterDirDown;
        else
            dir = eMonsterDirUp;

        // initialise chaque monstre du groupe
        for (int i = 0; i < 4; ++i)
            if (mMonsters[i].pos != eMonsterPosNone)
            {
                mMonsters[i].dir = dir;
                mMonsters[i].initNextAttack(type);
            }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::fight(CVec2 mapPos, uint8_t type)
{
    // transition en mode ballade ?
    CVec2   dist = player.pos - mapPos;
    bool    isNearPlayer = (dist.x == 0 && ABS(dist.y) == 1) || (dist.y == 0 && ABS(dist.x) == 1);

    if (isNearPlayer == false)
    {
        state = eMonsterState_Wandering;
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::hitMonster(CVec2 groupPos, int monsterNum, int damages)
{
    mMonsters[monsterNum].life -= damages;

    // mort ?
    if (mMonsters[monsterNum].life <= 0)
    {
        mMonsters[monsterNum].pos = eMonsterPosNone;

        // si tous les monstres sont morts, on supprime le groupe
        int i;
        for (i = 0; i < 4; ++i)
        {
            if (mMonsters[i].pos != eMonsterPosNone)
                break;
        }

        if (i == 4)
        {
            int groupIndex = map.findMonsterGroupIndex(groupPos);
            map.removeMonsterGroup(groupPos);
            monsters.monsterGroups.erase(monsters.monsterGroups.begin() + groupIndex);
        }
    }
}
