﻿#ifndef MONSTERS_H
#define MONSTERS_H

#include <QString>
#include <QImage>
#include <vector>
#include "../common/sources/system/cmaths.h"
#include "timer.h"

//---------------------------------------------------------------------------------------------
enum EMonsterPos
{
    eMonsterPosNone,    // empty monster
    eMonsterPosNorthWest,
    eMonsterPosNorthEast,
    eMonsterPosSouthWest,
    eMonsterPosSouthEast,
    eMonsterPosCenter
};

enum EMonsterDir
{
    eMonsterDirUp,
    eMonsterDirDown,
    eMonsterDirLeft,
    eMonsterDirRight
};

enum EMonsterState
{
    eMonsterState_Wandering,
    eMonsterState_Fight
};

class CMonsterGroup2;

//---------------------------------------------------------------------------------------------
class CMonster
{
public:
    void    update(CMonsterGroup2& group, CVec2 mapPos, uint8_t type, bool isInFront);
    void    wander();
    void    fight(CVec2 mapPos, uint8_t type, bool isInFront);
    void    initNextAttack(uint8_t type);
    uint8_t chooseTarget(CVec2 mapPos);

    EMonsterPos pos;
    EMonsterDir dir;
    int life;
    CTimer  displayAttackTimer;
    CTimer  nextAttackTimer;
};

//---------------------------------------------------------------------------------------------
class CMonsterGroup2
{
public:
    CMonsterGroup2();
    void    update(CVec2 mapPos, uint8_t type);
    void    wander(CVec2 mapPos, uint8_t type);
    void    fight(CVec2 mapPos, uint8_t type);
    void    hitMonster(CVec2 groupPos, int monsterNum, int damages);

    uint8_t getMonsterRow(CVec2 mapPos, EMonsterPos monsterPos);
    bool    isMonsterInFront(CVec2 mapPos, EMonsterPos monsterPos);

    CMonster        mMonsters[4];
    EMonsterDir     dir;
    EMonsterState   state;
};

//---------------------------------------------------------------------------------------------
class CMonsters
{
public:
    struct SMonsterData
    {
        QString frontImage;
        QString sideImage;
        QString backImage;
        QString attackImage;
        QString attackSound;
        int     attackDisplayTime;
        int     timeBetweenAttack;
        int     defense;
        int     baseHealth;
        int     attack;
        int     dexterity;
    };

    CMonsters();
    void    readMonstersDB();
    void    init();
    void    draw(QImage* image, CVec2 mapPos, CVec2 tablePos);
    void    update();

    std::vector<SMonsterData>   monstersDatas;
    std::vector<CMonsterGroup2> monsterGroups;

private:
    CVec2   getMonsterPos(float x, float y);
    int     getStartingLife(int type);
};

extern CMonsters    monsters;

#endif // MONSTERS_H
