﻿#include "mouse.h"
#include "bridge.h"
#include "../common/sources/system/cgraphics2d.h"
#include "../common/sources/system/filecache.h"
#include "objects.h"
#include "interface.h"

CMouse  mouse;

//-----------------------------------------------------------------------------------------
CMouse::CMouse()
{
    mPos = CVec2();
    mButtonPressing = false;
    mButtonReleasing = false;
    mButtonPressed = false;
    mObjectInHand.setType(0);
}

//-----------------------------------------------------------------------------------------
void CMouse::update()
{
    mPos = bridge.mMousePos;
    mButtonPressing = (mButtonPressed == false && bridge.mMouseButton == true);
    mButtonReleasing = (mButtonPressed == true && bridge.mMouseButton == false);
    mButtonPressed = bridge.mMouseButton;
}

//-----------------------------------------------------------------------------------------
void CMouse::clearAreas()
{
    mAreaList.clear();
}

//-----------------------------------------------------------------------------------------
void CMouse::addArea(EMouseAreaTypes type, CRect rect, ECursorTypes cursor, void* param1, void* param2, void *param3)
{
    SMouseArea  newArea;
    newArea.type = type;
    newArea.rect = rect;
    newArea.param1 = param1;
    newArea.param2 = param2;
    newArea.param3 = param3;
    newArea.cursor = cursor;

    mAreaList.push_back(newArea);
}

//-----------------------------------------------------------------------------------------
SMouseArea* CMouse::clickedArea(QImage* image)
{
    SMouseArea*  area = NULL;

    for (int i = mAreaList.size() - 1; i >= 0; --i)
    {
        CRect*  rect = &mAreaList[i].rect;

        if (mPos.x >= rect->tl.x &&
            mPos.y >= rect->tl.y &&
            mPos.x <= rect->br.x &&
            mPos.y <= rect->br.y)
        {
            area = &mAreaList[i];
            break;
        }
    }

    if (area == NULL || area->cursor != eCursor_None)
    {
        if (area == NULL)
        {
            QImage  cursor = fileCache.getImage("gfx/interface/CursorArrow.png");
            graph2D.drawImage(image, mPos, cursor);
        }
        else if (area->cursor == eCursor_Arrow)
        {
            QImage  cursor = fileCache.getImage("gfx/interface/CursorArrow.png");
            graph2D.drawImage(image, mPos, cursor);
        }
        else
        {
            if (mObjectInHand.getType() == 0)
            {
                QImage  cursor = fileCache.getImage("gfx/interface/CursorHand.png");
                graph2D.drawImage(image, mPos, cursor);
            }
            else
            {
                CObjects::CObjectInfo  object = objects.mObjectInfos[mObjectInHand.getType()];
                QImage  cursor = fileCache.getImage(object.imageFile.toLocal8Bit().constData());
                CRect   rect = interface.getItemRect(object.imageNum);
                CVec2   pos = mPos - CVec2(ITEM_WIDTH / 2, ITEM_HEIGHT / 2);
                graph2D.drawImageAtlas(image, pos + CVec2(2, 2), cursor, rect, BLACK);
                graph2D.drawImageAtlas(image, pos, cursor, rect);
            }
        }

    }

    if (mButtonPressed == true)
    {
        return area;
    }
    return NULL;
}
