﻿#include "objects.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "interface.h"
#include "player.h"
#include "walls.h"
#include "game.h"

CObjects objects;

CObjects::CObjects()
{

}

//---------------------------------------------------------------------------------------------
void    CObjects::readObjectsDB()
{
    QDomDocument doc;

    QFile f("databases/items.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    itemNode = root.firstChild();

    while(!itemNode.isNull())
    {
        if (itemNode.isElement())
        {
            QDomElement item = itemNode.toElement();
            std::vector<CParamType> paramsTypesList;

            if (item.tagName() == "item")
            {
                CObjectInfo object;
                object.name = interface.getText(item.attribute("name").toLocal8Bit().constData());
                object.combo = 0;
                object.damages = 0;
                QDomElement itemInfo = item.firstChild().toElement();

                while(!itemInfo.isNull())
                {
                    if (itemInfo.tagName() == "category")
                    {
                        object.category = itemInfo.text();
                    }
                    else if (itemInfo.tagName() == "floor_image")
                    {
                        object.floorImage = QString("gfx/3DView/items_floor/") + itemInfo.text();
                    }
                    else if (itemInfo.tagName() == "image_file")
                    {
                        object.imageFile = QString("gfx/interface/") + itemInfo.text();
                    }
                    else if (itemInfo.tagName() == "image_num")
                    {
                        object.imageNum = itemInfo.text().toInt();
                    }
                    else if (itemInfo.tagName() == "position")
                    {
                        object.positions = itemInfo.text();
                    }
                    else if (itemInfo.tagName() == "weight")
                    {
                        object.weight = itemInfo.text().toInt();
                    }
                    else if (itemInfo.tagName() == "combo")
                    {
                        object.combo = itemInfo.text().toInt();
                    }
                    else if (itemInfo.tagName() == "damages")
                    {
                        object.damages = itemInfo.text().toInt();
                    }
                    else if (itemInfo.tagName() == "param")
                    {
                        QString paramType = itemInfo.attribute("type");

                        CParamType  newParamType;

                        if (paramType == "string")
                        {
                            newParamType.mType = eParamString;
                            newParamType.mName = itemInfo.text();
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "int")
                        {
                            newParamType.mType = eParamInt;
                            newParamType.mName = itemInfo.text();
                            paramsTypesList.push_back(newParamType);
                        }
                    }

                    itemInfo = itemInfo.nextSibling().toElement();
                }

                objects.mObjectInfos.push_back(object);
                CMap::mObjectsParams.push_back(paramsTypesList);
            }
        }
        itemNode = itemNode.nextSibling();
    }
}

//---------------------------------------------------------------------------------------------
CVec2 CObjects::getObjectPos(CVec2 tablePos)
{
    CVec2   pos;
    pos.x = 250 * (tablePos.x - 4.5) / (8.5 - tablePos.y) + 112;
    pos.y = 310 / (8.5 - 0.9 * tablePos.y) + 66;
    return pos;
}

//---------------------------------------------------------------------------------------------
CVec2 CObjects::getAlcovePos(CVec2 tablePos)
{
    CVec2   pos;
    pos.x = 250 * (tablePos.x * 2 - 4) / (8.5 - tablePos.y * 2) + 112;
    static int  yTable[] = {91, 104, 120};
    pos.y = yTable[tablePos.y - 1];
    return pos;
}

//---------------------------------------------------------------------------------------------
bool CObjects::isWallInFront()
{
    CTile*  tile = map.getTile(player.pos);
    EWallSide   side = player.getWallSide(eWallSideUp);
    CWall*  wall = &tile->mWalls[side];

    return (wall->getType() != eWallNothing);
}

//---------------------------------------------------------------------------------------------
void CObjects::drawObjectsStack(QImage* image, CVec2 mapPos, EWallSide side, CVec2 tablePos)
{
    if (tablePos.y < 1 ||
        tablePos.y > (WALL_TABLE_HEIGHT - 1) * 2)
        return;

    CObjectStack*   stack = map.findObjectsStack(mapPos, side);

    if (stack != NULL)
    {
        for (size_t i = 0; i < stack->getSize(); ++i)
        {
            int type = stack->getObject(i).getType();

            if (type != 0)
            {
                // récupère l'image de l'objet
                CObjectInfo object = mObjectInfos[type];
                QImage  objectImage = fileCache.getImage(object.floorImage.toLocal8Bit().constData());

                // récupère la position de l'objet et ajoute une valeur pseudo-aléatoire
                CVec2   pos;

                if (side == eWallSideMax)
                    pos = getObjectPos(tablePos);
                else
                    pos = getAlcovePos(tablePos);
                pos.x += ((mapPos.x - mapPos.y + i) * 53) % 7 - 3;
                pos.y += (i < 4 ? i : 3);

                // calcule la taille par rapport à la position de référence (la plus proche)
                CVec2   pos0, pos1;
                float   scale;

                if (side == eWallSideMax)
                {
                    pos0 = getObjectPos(CVec2(WALL_TABLE_WIDTH, (WALL_TABLE_HEIGHT - 1) * 2));
                    pos1 = getObjectPos(CVec2(WALL_TABLE_WIDTH, tablePos.y));
                    scale = (float)(pos1.x - 112) / (float)(pos0.x - 112);
                }
                else
                {
                    static float scalesTable[] = {0.36, 0.53, 0.82};
                    scale = scalesTable[tablePos.y - 1];
                }

                // redimensionne l'objet dans une image temporaire
                QImage  scaledObject(objectImage.size() * scale, QImage::Format_ARGB32);
                scaledObject.fill(TRANSPARENT);
                graph2D.drawImageScaled(&scaledObject, CVec2(), objectImage, scale);

                // assombrit l'objet en fonction de sa distance
                float shadow;

                if (side == eWallSideMax)
                    shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePos.y) * 0.13f;
                else
                    shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePos.y * 2) * 0.13f;
                graph2D.darken(&scaledObject, shadow);

                // dessine l'objet à l'écran
                pos -= CVec2(scaledObject.size().width() / 2, scaledObject.size().height() - 1);
                graph2D.drawImage(image, pos, scaledObject, 0, false, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));

                // ajoute la zone souris
                if (mouse.mObjectInHand.getType() == 0)
                {
                    bool    isAreaOK = false;

                    if (side == eWallSideMax)
                    {
                        if (tablePos.x >= WALL_TABLE_WIDTH - 1 && tablePos.x <= WALL_TABLE_WIDTH)
                        {
                            if (tablePos.y == (WALL_TABLE_HEIGHT - 1) * 2 ||
                                (tablePos.y == (WALL_TABLE_HEIGHT - 2) * 2 + 1 && isWallInFront() == false))
                            {
                                isAreaOK = true;
                            }
                        }
                    }
                    else
                    {
                        if (mapPos != walls.resurrectWallPos || side != walls.resurrectWallSide)
                            if (tablePos.x == WALL_TABLE_WIDTH / 2 && tablePos.y == WALL_TABLE_HEIGHT - 1)
                                isAreaOK = true;
                    }

                    if (isAreaOK)
                    {
                        CRect   mouseRect(pos, CVec2(pos.x + scaledObject.width() - 1, pos.y + scaledObject.height() - 1));
                        mouse.addArea(eMouseAreaG_PickObject, mouseRect, eCursor_Hand, (void*)stack, (void*)i);
                    }
                }
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
void CObjects::drawBackRow(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    CVec2   pos1, pos2;

    switch (player.dir)
    {
        case 0: // haut
            pos1 = CVec2(0, 0);
            pos2 = CVec2(1, 0);
            break;

        case 1: // gauche
            pos1 = CVec2(0, 1);
            pos2 = CVec2(0, 0);
            break;

        case 2: // bas
            pos1 = CVec2(1, 1);
            pos2 = CVec2(0, 1);
            break;

        case 3: // droite
            pos1 = CVec2(1, 0);
            pos2 = CVec2(1, 1);
            break;
    }

    CVec2   mapLeft = mapPos * 2 + pos1;
    CVec2   mapRight = mapPos * 2 + pos2;
    drawObjectsStack(image, mapLeft,  eWallSideMax, tablePos * 2 + CVec2(0, 0));
    drawObjectsStack(image, mapRight, eWallSideMax, tablePos * 2 + CVec2(1, 0));

    // zones souris pour poser un objet
    if (mouse.mObjectInHand.getType() != 0)
    {
        if (tablePos.x == WALL_TABLE_WIDTH / 2 &&
            tablePos.y == WALL_TABLE_HEIGHT - 1)
        {
            CRect   rectLeft(CVec2(31, 153), CVec2(111, 168));
            mouse.addArea(eMouseAreaG_DropObject, rectLeft, eCursor_Hand, (void*)mapLeft.x, (void*)mapLeft.y, (void*)eWallSideMax);
            CRect   rectRight(CVec2(112, 153), CVec2(192, 168));
            mouse.addArea(eMouseAreaG_DropObject, rectRight, eCursor_Hand, (void*)mapRight.x, (void*)mapRight.y, (void*)eWallSideMax);
        }
    }
}

//---------------------------------------------------------------------------------------------
void CObjects::drawFrontRow(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    CVec2   pos1, pos2;

    switch (player.dir)
    {
        case 0: // haut
            pos1 = CVec2(0, 1);
            pos2 = CVec2(1, 1);
            break;

        case 1: // gauche
            pos1 = CVec2(1, 1);
            pos2 = CVec2(1, 0);
            break;

        case 2: // bas
            pos1 = CVec2(1, 0);
            pos2 = CVec2(0, 0);
            break;

        case 3: // droite
            pos1 = CVec2(0, 0);
            pos2 = CVec2(0, 1);
            break;
    }

    CVec2   mapLeft = mapPos * 2 + pos1;
    CVec2   mapRight = mapPos * 2 + pos2;
    drawObjectsStack(image, mapLeft,  eWallSideMax, tablePos * 2 + CVec2(0, 1));
    drawObjectsStack(image, mapRight, eWallSideMax, tablePos * 2 + CVec2(1, 1));

    // zones souris pour poser un objet
    if (mouse.mObjectInHand.getType() != 0)
    {
        if (tablePos.x == WALL_TABLE_WIDTH / 2 &&
            tablePos.y == WALL_TABLE_HEIGHT - 2 &&
            isWallInFront() == false)
        {
            CRect   rectLeft(CVec2(46, 138), CVec2(111, 152));
            mouse.addArea(eMouseAreaG_DropObject, rectLeft, eCursor_Hand, (void*)mapLeft.x, (void*)mapLeft.y, (void*)eWallSideMax);
            CRect   rectRight(CVec2(112, 138), CVec2(177, 152));
            mouse.addArea(eMouseAreaG_DropObject, rectRight, eCursor_Hand, (void*)mapRight.x, (void*)mapRight.y, (void*)eWallSideMax);
        }
    }
}

//---------------------------------------------------------------------------------------------
void CObjects::setDefaultParams(CObject* obj)
{
    if (obj->getType() == OBJECT_TYPE_TORCH)
    {
        obj->setIntParam("Charges", TORCH_MAX_CHARGES);
    }
}
